#ifndef OPTION_PARSER_UTIL_H
#define OPTION_PARSER_UTIL_H

/*
  TODO: This file was intended to be a temporary stub. We should think
  of getting rid of it. This would entail including "options/options.h"
  instead and explicitly using the options namespace. See also option_parser.h,
  plugin.h and issue588 for a discussion.
*/
#include "options/bounds.h"
#include "options/options.h"

using options::Bounds;
using options::Options;
using options::ParseTree;

#endif
