#include "types.h"

#include <limits>

using namespace std;

namespace merge_and_shrink {
const int INF = numeric_limits<int>::max();
const int MINUSINF = numeric_limits<int>::min();
const int PRUNED_STATE = -1;
}
